/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#include <stdio.h>
#include "main.h"
#include "mem.h"
#include "inst.h"
#include "smt.h"

extern unsigned long cyc;
extern context *thecontexts[];

/* This routine prints out a disassembly-like listing for
 *  debugging.
 */
    
extern void print(int thread, unsigned long pc, unsigned steps) {
    unsigned int inst;
    unsigned int offset, ra, rb, rc; /* for decoding */
    unsigned long rbv;
    unsigned long va;
    int i;

    context *current = thecontexts[thread];

    for (i=0;i<thread;i++) printf("\t");
    printf("T%ds%d:",thread,current->scoretop);
    if (!current->wrong_path && !current->follow_sync) 
      printf("C:");
    else {
      if (current->wrong_path) {
	printf("W:");
	if (!pc) {
	  printf(":\t<%ld>PC = 0\n", cyc);
	  return;
	}
      }
      if (current->follow_sync) printf("S:");
    }
    while(steps-- > 0) {
        inst = rnr_mem_word(current, pc);
	printf("0x%0lx", pc);
	printf(":\t<%ld>", cyc);
	printf(":\t[0x%08x]\t", inst); 
	ra = (inst >> 21) & 0x1f;
	rb = (inst >> 16) & 0x1f;
	rc = inst & 0x1f;
	offset = inst & 0xffff;
	switch(inst >> 26) {
	case LDA:printf("lda"); goto mem;
	case LDAH:printf("ldah"); goto mem;
	case LDQ_U:printf("ldq_u"); goto memlq;
	case STQ_U:printf("stq_u"); goto memsq;
	case INTA:
	  switch (offset >> 5 & 0x7f) {
	  case ADDLV:printf("addl/v"); goto reg3;
	  case ADDL:printf("addl"); goto reg3;
	  case ADDQ:printf("addq"); goto reg3;
	  case ADDQV:printf("addq/v"); goto reg3;
	  case CMPLE:printf("cmple"); goto reg3;
	  case CMPULT:printf("cmpult"); goto reg3;
	  case S4ADDL:printf("s4addl"); goto reg3;
	  case S4SUBQ:printf("s4subq"); goto reg3;
	  case S8SUBL:printf("s8subl"); goto reg3;
	  case CMPBGE:printf("cmpbge"); goto reg3;
	  case CMPLT:printf("cmplt"); goto reg3;
	  case SUBLV:printf("sublv"); goto reg3;
	  case SUBL:printf("subl"); goto reg3;
	  case SUBQV:printf("subq/v"); goto reg3;
	  case SUBQ:printf("subq"); goto reg3;
	  case S4ADDQ:printf("s4addq"); goto reg3;
	  case S8ADDL:printf("s8addl"); goto reg3;
	  case S8SUBQ:printf("s8subq"); goto reg3;
	  case CMPEQ:printf("cmpeq"); goto reg3;
	  case CMPULE:printf("cmpule"); goto reg3;
	  case S4SUBL:printf("s4subl"); goto reg3;
	  case S8ADDQ:printf("s8addq"); goto reg3;
	  reg3: 
	    if (inst & 0x1000) {
	      rbv = (inst >> 13) & 0xff;
	      printf(" $%d,%d, $%d\n", ra, rbv, rc);
	    }
	    else {
	      printf(" $%d,$%d, $%d\n", ra, rb, rc);
	    }
	    break;
	  }
	  break;
	case INTL:
	  switch (offset >>5 & 0x7f) {
	  case AND:printf("and"); goto reg3_;
	  case BIC:printf("bic"); goto reg3_;
	  case BIS:
	    switch (inst) {
	    case SMT_FORK:
	      printf("smt_fork\n");break;
	    case SMT_JOIN:
	      printf("smt_join\n");break;
	    case SMT_GETID:
	      printf("smt_getid\n");break;
	    case SMT_CREATESTACK:
	      printf("smt_createstack\n");break;
	    case SMT_PRINT_INT:
	      printf("smt_print_int\n");break;
	    case SMT_TAS:
	      printf("smt_tas\n");break;
	    case SMT_TASQ:
	      printf("smt_tasq\n");break;
	    case SMT_STOP:
	      printf("smt_stop\n");break;
	    case SMT_RESTART:
	      printf("smt_restart\n");break;
	    case SMT_HW_LOCK:
	      printf("smt_HW_LOCK\n");break;
	    case SMT_RELEASE:
	      printf("smt_RELEASE\n");break;
	    default:
	      printf("bis"); goto reg3_;
	    }
	    break;
	  case CMOVEQ:printf("cmoveq"); goto reg3_;
	  case CMOVLBC:printf("cmovlbc"); goto reg3_;
	  case CMOVLBS:printf("cmovlbs"); goto reg3_;
	  case CMOVGE:printf("cmovge"); goto reg3_;
	  case CMOVGT:printf("cmovgt"); goto reg3_;
	  case CMOVLE:printf("cmovle"); goto reg3_;
	  case CMOVLT:printf("cmovlt"); goto reg3_;
	  case CMOVNE:printf("cmovne"); goto reg3_;
	  case EQV:printf("eqv"); goto reg3_;
	  case ORNOT:printf("ornot"); goto reg3_;
	  case XOR:printf("xor"); goto reg3_;
	  reg3_: 
	    if (inst & 0x1000) {
	      rbv = (inst >> 13) & 0xff;
	      printf(" $%d,%d, $%d\n", ra, rbv, rc);
	    }
	    else {
	      printf(" $%d,$%d, $%d\n", ra, rb, rc);
	    }
	    break;
	  }
	  break;
	case INTS:
	  switch (offset >>5 & 0x7f) {
	  case EXTBL:printf("extbl"); goto reg3__;
	  case EXTLH:printf("extlh"); goto reg3__;
	  case EXTLL:printf("extll"); goto reg3__;
	  case EXTQH:printf("extqh"); goto reg3__;
	  case EXTQL:printf("extql"); goto reg3__;
	  case EXTWH:printf("extwh"); goto reg3__;
	  case EXTWL:printf("extwl"); goto reg3__;
	  case INSBL:printf("insbl"); goto reg3__;
	  case INSLH:printf("inslh"); goto reg3__;
	  case INSLL:printf("insll"); goto reg3__;
	  case INSQH:printf("insqh"); goto reg3__;
	  case INSQL:printf("insql"); goto reg3__;
	  case INSWH:printf("inswh"); goto reg3__;
	  case INSWL:printf("inswl"); goto reg3__;
	  case MSKBL:printf("mskbl"); goto reg3__;
	  case MSKLH:printf("msklh"); goto reg3__;
	  case MSKLL:printf("mskll"); goto reg3__;
	  case MSKQH:printf("mskqh"); goto reg3__;
	  case MSKQL:printf("mskql"); goto reg3__;
	  case MSKWH:printf("mskwh"); goto reg3__;
	  case MSKWL:printf("mskwl"); goto reg3__;
	  case SLL:printf("sll"); goto reg3__;
	  case SRA:printf("sra"); goto reg3__;
	  case SRL:printf("srl"); goto reg3__;
	  case ZAP:printf("zap"); goto reg3__;
	  case ZAPNOT:printf("zapnot"); goto reg3__;
	  reg3__: 
	    if (inst & 0x1000) {
	      rbv = (inst >> 13) & 0xff;
	      printf(" $%d,%d, $%d\n", ra, rbv, rc);
	    }
	    else {
	      printf(" $%d,$%d, $%d\n", ra, rb, rc);
	    }
	    break;
	  }
	  break;
	case INTM:
	  switch (offset >>5 & 0x7f) {
	  case MULLV:printf("mull/v"); goto reg3___;
	  case MULL:printf("mull"); goto reg3___;
	  case MULQV:printf("mulq/v"); goto reg3___;
	  case MULQ:printf("mulq"); goto reg3___;
	  case UMULH:printf("umulh"); goto reg3___;
	  reg3___: 
	    if (inst & 0x1000) {
	      rbv = (inst >> 13) & 0xff;
	      printf(" $%d,%d, $%d\n", ra, rbv, rc);
	    }
	    else {
	      printf(" $%d,$%d, $%d\n", ra, rb, rc);
	    }
	    break;
	  }
	  break;
	case FLTL:
	  switch (offset >>5) {
	  case CPYS:printf("cpys");goto _reg3;
	  case CPYSE:printf("cpyse");goto _reg3;
	  case CPYSN:printf("cpysn");goto _reg3;
	  case CVTLQ:printf("cvtlq");goto reg2_;
	  case CVTQL:printf("cvtql");goto reg2_;
	  case CVTQLSV:printf("cvtql/sv");goto reg2_;
	  case CVTQLV:printf("cvtql/v");goto reg2_;
	  case FCMOVEQ:printf("fcmoveq");goto _reg3;
	  case FCMOVGE:printf("fcmovege");goto _reg3;
	  case FCMOVGT:printf("fcmovgt");goto _reg3;
	  case FCMOVLE:printf("fcmovle");goto _reg3;
	  case FCMOVLT:printf("fcmovlt");goto _reg3;
	  case FCMOVNE:printf("fcmovne");goto _reg3;
	  case MT_FPCR:printf("mt_fpcr %d", ra);break;
	  case MF_FPCR:printf("mf_fpcr %d", ra);break;
	  reg2_:
	    printf(" $%d, $%d\n", rb, rc);
	    break;
	  _reg3: 
	    printf(" $%d,$%d, $%d\n", ra, rb, rc);
	    break;
	  }
	  break;
	case FLTI:
	  switch ((offset >>5) & 0x3f) {
	  case ADDS:printf("adds");goto __reg3;
	  case ADDT:printf("addt");goto __reg3;
	  case CMPTEQ:printf("cmpteq");goto __reg3;
	  case CMPTLT:printf("cmptlt");goto __reg3;
	  case CMPTLE:printf("cmptle");goto __reg3;
	  case CMPTUN:printf("cmptun");goto __reg3;
	  case CVTQS:printf("cvtqs");goto __reg3;
	  case CVTQT:printf("cvtqt");goto __reg3;
	  case CVTTS:printf("cvtts");goto __reg3;
	  case DIVS: printf("divs");goto __reg3;
	  case DIVT:printf("divt");goto __reg3;
	  case MULS:printf("muls");goto __reg3;
	  case MULT:printf("mult");goto __reg3;
	  case SUBS:printf("subs");goto __reg3;
	  case SUBT:printf("subt");goto __reg3;
	  case CVTTQ:printf("cvttq");goto __reg3;
	  __reg3: 
	    printf(" $%d,$%d, $%d\n", ra, rb, rc);
	    break;
	  }
	  break;
	case MISC:
	  switch (offset) {
	  case FETCH:printf("fetch\n");break;
	  case RC : printf("rc\n");break;
	  case TRAPB: printf("trapb\n");break;
	  case MB: printf("mb\n");break;
	  case RS: printf("rs\n");break;
	  case FETCH_M: printf("fetch_m\n");break;
	  case RPCC: printf("rpcc\n");break;
	  }
	  break;
	case JMPJSR:printf("jmp $%d, ($%d)",ra, rb);
	  printf("[0x%lx]\n", current->R[rb] & (~0x11));
	  break;
	case LDS:printf("lds");goto memfs;
	case LDF:printf("ldf");goto memf;
	case LDT:printf("ldt");goto memft;
	case LDG:printf("ldg");goto memf;
 	case STS:printf("sts");goto memsfs;
	case STF:printf("stf");goto memsf;
	case STT:printf("stt");goto memsft;
	case STG:printf("stg");goto memsf;
	case LDL:printf("ldl");goto memlw;
	case LDQ:printf("ldq");goto memlq;
	case LDL_L:printf("ldl_l");goto memlw;
	case LDQ_L:printf("ldq_l");goto memlq;
	case STL:printf("stl");goto memsw;
	case STQ:printf("stq");goto memsq;
	case STL_C:printf("stl_c");goto memsw;
	case STQ_C:printf("stq_c");goto memsq;
	case BSR:printf("bsr");goto branch; 
	case BR:printf("br");goto branch; 
	case FBEQ:printf("fbeq");goto fbranch; 
	case FBLT:printf("fblt");goto fbranch; 
	case FBLE:printf("fble");goto fbranch; 
	case FBNE:printf("fbne");goto fbranch; 
	case FBGE:printf("fbge");goto fbranch; 
	case FBGT:printf("fbgt");goto fbranch; 
	case BLBC:printf("blbc");goto branch; 
	case BEQ:printf("beq");goto branch; 
	case BLT:printf("blt");goto branch; 
	case BLE:printf("ble");goto branch; 
	case BLBS:printf("blbs");goto branch; 
	case BNE:printf("bne");goto branch; 
	case BGE:printf("bge");goto branch; 
	case BGT:printf("bgt");goto branch; 
	case CALL_PAL:
	  switch(inst & 0x3ffffff) {
	  case callsys:printf("syscall\n");break;
	  case rduniq:printf("rduniq\n");break;
	  case wruniq:printf("wruniq\n");break;
	  default:printf("call_pal\n");
	  }
	  break;
	mem:
	  va = current->R[rb]+SIGNEX(offset);
	  printf(" $%d, %d($%d) [0x%lx]\n",ra,SIGNEX(offset),rb, va);
	  break;
	memf:
	  va = current->R[rb]+SIGNEX(offset);
	  printf(" $f%d, %d($%d) [0x%lx]\n",ra,SIGNEX(offset),rb, va);
	  break;
	memfs:
	  va = current->R[rb]+SIGNEX(offset);
	  printf(" $f%d, %d($%d) [0x%lx] = %.2f\n",ra,SIGNEX(offset),rb, va,
		 current->wrong_path?0:read_mem_word(current, va));
	  break;
	memft:
	  va = current->R[rb]+SIGNEX(offset);
	  printf(" $f%d, %d($%d) [0x%lx]\n",ra,SIGNEX(offset),rb, va);
	  break;
	memlw:
	  va = current->R[rb]+SIGNEX(offset);
	  printf(" $%d, %d($%d) [0x%lx] = %lx\n",ra,SIGNEX(offset),rb,
		 va, (current->wrong_path||current->follow_sync)?0:read_mem_word(current, va));
	  break;
	memsw:
	  va = current->R[rb]+SIGNEX(offset);
	  printf(" $%d, %d($%d) [0x%lx] = 0x%lx\n",ra,SIGNEX(offset),rb,
		 va, current->R[ra]);
	  break;
	memsf:
	  va = current->R[rb]+SIGNEX(offset);
	  printf(" $f%d, %d($%d) [0x%lx] = %.2f\n",ra,SIGNEX(offset),rb,
		 va, current->FWR[ra]);
	  break;
	memsfs:
	  va = current->R[rb]+SIGNEX(offset);
	  printf(" $f%d, %d($%d) [0x%lx] = %.2f\n",ra,SIGNEX(offset),rb,
		 va, 
		 (((current->FWR[ra]>>32)&0xc0000000)|
		  ((current->FWR[ra]>>29) & 0x3ffffff)));
	  break;
	memsft:
	  va = current->R[rb]+SIGNEX(offset);
	  printf(" $f%d, %d($%d) [0x%lx] = %.2f\n",ra,SIGNEX(offset),rb,
		 va, current->FWR[ra]);
	  break;
	memlq:
	  va = current->R[rb]+SIGNEX(offset);
/*	  printf(" $%d, %d($%d) [0x%lx] = 0x%lx\n",ra,SIGNEX(offset),rb,
		 va, read_mem_quad(current, va));*/
	  printf(" $%d, %d($%d) [0x%lx] = %ld\n",ra,SIGNEX(offset),rb,va,
		 (current->wrong_path||current->follow_sync)?0:read_mem_quad(current, va));
	  break;
	memsq:
	  va = current->R[rb]+SIGNEX(offset);
	  printf(" $%d, %d($%d) [0x%lx] = 0x%lx\n",ra,SIGNEX(offset),rb,va,
		 current->R[ra]);
	  break;
	branch:
	  printf(" $%d [0x%lx], ",ra, current->R[ra]);
	  printf("0x%0lx\n", pc + (SEXT(offset) << 2) + 4);
	  break;
	fbranch:
	  printf(" $%d [%.2f], ",ra, current->FWR[ra]);
	  printf("0x%0lx\n", pc + (SEXT(offset) << 2) + 4);
	  break;
	reg2:
	  printf(" $%d, $%d\n", rb, ra);
	  break;
	}
	pc += 4;
    }
}
