/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#include "stash.h"
#include "main.h"
#include "queue.h"
#include "mem.h"
#include "cache.h"
#include "fetch.h"
#include "inst.h"
#include "smt.h"

#define FUTYPES 4
#define MAXINT 6
#define MAXFP 3
#define MAXLDST 4
#define MAXSYNCH 2

scoreboard_t *iq_head;
scoreboard_t *iq_tail;
scoreboard_t *fq_head;
scoreboard_t *fq_tail;

extern scoreboard_t *rread_head;
extern scoreboard_t *rread_tail;

extern int num_in_rread;
extern unsigned long cyc;

extern context *thecontexts[];
extern int numcontexts;
extern int key[];
extern int debugsync;

int numiniq, numinfq;
unsigned long instrcount[FUTYPES] = {0,0,0,0};
unsigned long wpexec, loadiqocc=0, loadfqocc=0;


extern unsigned long flushed, execflushed;
extern unsigned long misfetchtotal;
extern long iqsizetotal;
extern long fqsizetotal;
extern int iqconf, fqconf, iregconf, fregconf;
extern unsigned long scoreconf;
static unsigned long intfuconf=0, fpfuconf=0, ldstfuconf=0;

long oldtotal, oldinstructions[MAXTHREADS], oldcycles;
int resetflag = 0;

/*  It turns out that most of the important action takes place
 *    here, so that the most important stats are both recorded
 *    in and printed out from this file.  
 *  The queue stage is when instructions are selected for issue
 *    from the instruction queues, then all resulting dependences
 *    are resolved to possibly wake up other instructions for
 *    execution. 
 */
 
extern void
schedstats(void)
{
  long totalcycles = TOTALFETCHLIMIT * cyc;
  long total=0, wptotal=0;
  int i;
  long totalexec;

    totalexec = 0;
    for (i=0;i<4;i++) {
      totalexec += instrcount[i];
    }
    printf("Threads: %d\t\tFetch Width: %d\n", numcontexts, TOTALFETCHLIMIT);
    for (i=0;i<numcontexts;i++) { 
      total += thecontexts[i]->instrs;
      wptotal += thecontexts[i]->wpinstrs;
    }

    printf("No of useful instructions fetched:\t%ld (%.2f%% proc util.)\n", total,
	   (float) 100.0 * total / totalcycles);
 printf("No of total instructions fetched:\t%ld (%.2f%% proc util.)\n", 
	(wptotal+total+misfetchtotal+flushed), 
	(float) 100.0 * (wptotal+total+misfetchtotal+flushed) / totalcycles);
 printf(" includes %d misfetch (%.1f%%), %d (%.1f%%) wrong-path fetches\n and %d flushed after failed lock(%.1f%%)\n",
	misfetchtotal,
	(float)100.0 * ((float)misfetchtotal/(misfetchtotal+wptotal+flushed+total)),
	wptotal,
	(float)100.0 * ((float)wptotal/(misfetchtotal+wptotal+flushed+total)),
	flushed,
	(float)100.0 * ((float)flushed/(misfetchtotal+wptotal+flushed+total)));
    printf("No of useful instructions issued/executed:\t%ld (%.2f%% proc util.)\n", 
	   totalexec-wpexec-execflushed, 
	   (float) 100.0 * (totalexec-wpexec-execflushed) / totalcycles);
    printf("No of total instructions issued/executed:\t%ld (%.2f%% proc util.)\n", 
	   (totalexec), (float) 100.0 * (totalexec) / totalcycles);
    printf(" includes %d wrong-path (%.1f%%) insts and %d flushed (%.1f%%)\n",
	   wpexec, (float) 100.0 * ((float)wpexec/(totalexec)),
	   execflushed, (float) 100.0 * ((float)execflushed/(totalexec)));
    printf("\tfp %ld, %.2f%%\n",instrcount[0],
	   (float) 100.0 * (float) instrcount[0]/(float) totalexec);
    printf("\tsynch %ld, %.2f%%\n",instrcount[1],
	   (float) 100.0 * (float) instrcount[1]/(float) totalexec);
    printf("\tinteger %ld, %.2f%%\n",instrcount[2],
	   (float) 100.0 * (float) instrcount[2]/(float) totalexec);
    printf("\tload/store %ld, %.2f%%\n",instrcount[3],
	   (float) 100.0 * (float) instrcount[3]/(float) totalexec);
    printf("Avg size of int IQ (before scheduling): %.2f\n",
	   (float) iqsizetotal/cyc);
    printf("Avg size of fp IQ (before scheduling): %.2f\n",
	   (float) fqsizetotal/cyc);
    printf("No of cycles taken:\t\t%lu\n", cyc);
    printf("Parallelism [");
    total = 0;
    for (i=0;i<numcontexts;i++) {
      total += thecontexts[i]->instrs;
      printf(" %.3f,", (double) thecontexts[i]->instrs / cyc );
    }
  printf("] = %.3f\n", (double) total / cyc);
  if (resetflag) {
    printf("Cycles since print_sim:\t\t%lu\n", cyc-oldcycles);
    printf("Paral. since print_sim [");
    for (i=0;i<numcontexts;i++) {
      printf(" %.3f,", (double) (thecontexts[i]->instrs-oldinstructions[i]) /
	     (cyc - oldcycles) );
    }
    printf("] = %.3f\n", (double) (total-oldtotal) / (cyc-oldcycles));
  }
  printf(" IQ conflicts: %d (%.2f%%), FQ conflicts: %d (%.2f%%) \n",
	 iqconf, 100.0 * (float) iqconf/cyc, fqconf, 
	 100.0 * (float) fqconf/cyc);
  printf(" I reg conflicts: %d (%.2f%%), FP reg conflicts: %d (%.2f%%)\n", 
	 iregconf, 100.0 * (float) iregconf/cyc, fregconf, 
	 100.0 * (float) fregconf/cyc);
  printf("  Scoreboard conflicts: %d (%.2f%%)\n", 
	 scoreconf, 100.0 * (float) scoreconf/cyc);
  printf("  Int unit conflicts: %d (%.2f%%), FP unit conflicts: %d (%.2f%%)\n", 
	 intfuconf, 100.0 * (float) intfuconf/cyc, fpfuconf, 
	 100.0 * (float) fpfuconf/cyc);
  printf("  Load/store unit conflicts: %d (%.2f%%)\n", 
	 ldstfuconf, 100.0 * (float) ldstfuconf/cyc);
}

extern void
reset_stats(void)
{
  int i;

  oldtotal = 0;
  for (i=0;i<numcontexts;i++) { 
    oldtotal += thecontexts[i]->instrs;
    oldinstructions[i] = thecontexts[i]->instrs;
  }
  oldcycles = cyc;
  resetflag = 1;
}

static void
insert_in_rread(scoreboard_t *newinst)
{
  num_in_rread++;
  if (rread_head == NULL) {
    rread_head = rread_tail = newinst;
  }
  else {
    rread_tail->next = newinst;
    rread_tail = newinst;
  }
}

/*  There are several flavors of the resolve routine for various
 *    types of instructions.  We use the latency of this instruction
 *    to then figure out when dependent instructions could be
 *    issued, if indeed this is the last unresolved dependence.
 *  Of course, this is seriously complicated by the fact that we
 *    continue to track and obey dependences even when we are
 *    executing down wrong paths.
 */

static void resolve(struct scoreboard *exinst)
{
  int i,j;
  int exindex = exinst->id;
  long dmemdelay, dtlbdelay;
  struct scoreboard *waitinst;
  int index;
  context *current = thecontexts[exinst->thread];

  exinst->status = EXECUTING;

  if (exinst->srcmem) {
    dmemdelay = dodaccess(exinst->srcmem, C_READ, current, exinst);
#ifdef DEBUG
    if (dmemdelay) 
      if (exinst->status & MEMORY) 
	DEBUGPRINTF("T%ds%d data cache dtlb miss, R%d, read 0x%lx\n",
		    current->id, exinst->id, exinst->dest,
		    exinst->srcmem);
      else
	DEBUGPRINTF("T%ds%d data cache conflict %d cycles, R%d, read 0x%lx\n",
		    current->id, exinst->id, dmemdelay, exinst->dest,
		    exinst->srcmem);
#endif
    if (exinst->status & MEMORY) {
      exinst->donecycle = cyc + 10000;
      return;
    }
    if (exinst->wp || current->wrong_path) {
      exinst->donecycle = cyc + 
	exinst->delay + dmemdelay;
    }
    if (!exinst->wp) {
      exinst->donecycle = cyc + 
	exinst->delay + dmemdelay;
    }
  }
  else if (exinst->destmem) {
    exinst->donecycle = cyc + exinst->delay + 
      dodaccess(exinst->destmem,C_WRITE, current,exinst);
  }
  else if (exinst->dest == 31 || exinst->dest == 63) { /* no destination reg */
    exinst->donecycle = cyc + exinst->delay;
    return;
  }
  else { /* no memory operation */
    exinst->donecycle = cyc + exinst->delay;
  }
  if (exinst->wp || current->wrong_path) {
    if (current->writer_cp[exinst->dest] == exindex) {
      current->writer_cp[exinst->dest] = NONE;
      current->lasthazard_cp[exinst->dest] = exinst->donecycle;
    } 
  }
  if (current->follow_sync) {
    if (current->writer_sync_cp[exinst->dest] == exindex) {
      current->writer_sync_cp[exinst->dest] = NONE;
      current->lasthazard_sync_cp[exinst->dest] = exinst->donecycle;
    } 
  }
    if (current->writer[exinst->dest] == exindex) {
      current->writer[exinst->dest] = NONE;
      current->lasthazard[exinst->dest] = exinst->donecycle;
    }
  for (i=0;i<exinst->numwaiting;i++) {
    waitinst = exinst->waiter[i];
    exinst->waiter[i] = 0;
    if (waitinst->hazcycle < exinst->donecycle) {
      waitinst->hazcycle = exinst->donecycle;
    }
    if (--waitinst->deps == 0) {
      waitinst->status &= ~BLOCKED;
      waitinst->readycycle = waitinst->hazcycle;
    }
  }
  exinst->numwaiting = 0;
}

brresolve(scoreboard_t *exinst)
{
  int i,j;
  int exindex = exinst->id;
  struct scoreboard *waitinst;
  context *current = thecontexts[exinst->thread];

  exinst->status = EXECUTING;
  exinst->donecycle = cyc + exinst->delay;
  if (exinst->dest != 31 && exinst->dest != 63) {
    if (exinst->wp || current->wrong_path) {
      if (current->writer_cp[exinst->dest] == exindex) {
	current->writer_cp[exinst->dest] = NONE;
	current->lasthazard_cp[exinst->dest] = exinst->donecycle;
      }
    }
    if (current->follow_sync) {
      if (current->writer_sync_cp[exinst->dest] == exindex) {
	current->writer_sync_cp[exinst->dest] = NONE;
	current->lasthazard_sync_cp[exinst->dest] = exinst->donecycle;
      } 
    }
      if (current->writer[exinst->dest] == exindex) {
	current->writer[exinst->dest] = NONE;
	current->lasthazard[exinst->dest] = exinst->donecycle;
      }
  }
  for (i=0;i<exinst->numwaiting;i++) {
    waitinst = exinst->waiter[i];
    exinst->waiter[i] = 0;
    if (waitinst->hazcycle < exinst->donecycle) {
      waitinst->hazcycle = exinst->donecycle;
    }
    if (--waitinst->deps == 0) {
      waitinst->status &= ~BLOCKED;
      waitinst->readycycle = waitinst->hazcycle;
    }
  }
  exinst->numwaiting = 0;
}

synchresolve(struct scoreboard *exinst)
{
  int index, i;
  unsigned long dmemdelay, va;
  context *current = thecontexts[exinst->thread];
  struct scoreboard *waitinst;

  exinst->status = EXECUTING;
  if (exinst->srcmem) {
    dmemdelay = dodaccess(exinst->srcmem, C_READ, current, exinst);
#ifdef DEBUG
    if (dmemdelay) 
      if (exinst->status & MEMORY) 
	DEBUGPRINTF("T%ds%d synch data cache or tlb miss, R%d, read 0x%lx\n",
		    current->id, exinst->id, exinst->dest,
		    exinst->srcmem);
      else
	DEBUGPRINTF("T%ds%d synch data cache or tlb miss %d cycles, R%d, read 0x%lx\n",
		    current->id, exinst->id, dmemdelay, exinst->dest,
		    exinst->srcmem);
#endif
    if (exinst->status & MEMORY) {
      exinst->donecycle = cyc + 10000;
      return;
    }
    exinst->donecycle = cyc +	exinst->delay + dmemdelay;
    if (exinst->dest != 31 && exinst->dest != 63) {
      if (exinst->wp || current->wrong_path) {
	if (current->writer_cp[exinst->dest] == exinst->id) {
	  current->writer_cp[exinst->dest] = NONE;
	  current->lasthazard_cp[exinst->dest] = exinst->donecycle; 
	}
      }
      if (current->follow_sync) {
	if (current->writer_sync_cp[exinst->dest] == exinst->id) {
	  current->writer_sync_cp[exinst->dest] = NONE;
	  current->lasthazard_sync_cp[exinst->dest] = exinst->donecycle;
	} 
      }
	if (current->writer[exinst->dest] == exinst->id) {
	  current->writer[exinst->dest] = NONE;
	  current->lasthazard[exinst->dest] = exinst->donecycle;
	} 
    }
  }
  else if (exinst->destmem) {
    exinst->donecycle = cyc + exinst->delay +
      dodaccess(exinst->destmem,C_WRITE, current, exinst);
  }
  else
    exinst->donecycle = cyc + exinst->delay;
  if (exinst->syncop == LDL_L || exinst->syncop == LDQ_L) {
    for (i=0;i<exinst->numwaiting;i++) {
      waitinst = exinst->waiter[i];
      exinst->waiter[i] = 0;
      if (waitinst->hazcycle < exinst->donecycle) {
	waitinst->hazcycle = exinst->donecycle;
      }
      if (--waitinst->deps == 0) {
	waitinst->status &= ~BLOCKED;
	waitinst->readycycle = waitinst->hazcycle;
      }
    }
    exinst->numwaiting = 0;
  }
}

/* Unlike the other resolve routines which get called at issue, 
 *  the mem_resolve routines cannot execute until after the request
 *  has made its way through the memory hierarchy (for a miss),
 *  that is, through our memory event queues.
 */

synch_mem_resolve(struct scoreboard *exinst, long donetime)
{
  context *current = thecontexts[exinst->thread];
  int index, i;
  scoreboard_t *waitinst;

  DEBUGPRINTF("T%ds%d synch memory access complete at %ld\n",
	      exinst->thread, exinst->id, donetime + exinst->delay);

  exinst->status = EXECUTING;
  exinst->donecycle = donetime + exinst->delay;
  
  if (exinst->wp || current->wrong_path) {
    if (current->writer_cp[exinst->dest] == exinst->id) {/* FIXME --
							    should never happen
							    */
      printf("synch_resolve from wrong path????\n");
      current->writer_cp[exinst->dest] = NONE;
      current->lasthazard_cp[exinst->dest] = exinst->donecycle;
    } 
  }
  if (current->follow_sync) {
    if (current->writer_sync_cp[exinst->dest] == exinst->id) { /* FIXME */
      /*printf("synch_resolve from follow_sync????\n");*/
      current->writer_sync_cp[exinst->dest] = NONE;
      current->lasthazard_sync_cp[exinst->dest] = exinst->donecycle;
    } 
  }
  if (current->writer[exinst->dest] == exinst->id) {
    current->writer[exinst->dest] = NONE;
    current->lasthazard[exinst->dest] = exinst->donecycle;
  } 
  for (i=0;i<exinst->numwaiting;i++) {
    waitinst = exinst->waiter[i];
    exinst->waiter[i] = 0;
    if (waitinst->hazcycle < exinst->donecycle) {
      waitinst->hazcycle = exinst->donecycle;
    }
    if (--waitinst->deps == 0) {
      waitinst->status &= ~BLOCKED;
      waitinst->readycycle = waitinst->hazcycle;
    }
  }
  exinst->numwaiting = 0;
}


void mem_resolve(struct scoreboard *exinst, unsigned long donetime)
{
  long dmemdelay;
  struct scoreboard *waitinst;
  int index, i;
  context *current = thecontexts[exinst->thread];

  DEBUGPRINTF("T%ds%d memory access complete at %ld\n",
	      exinst->thread, exinst->id, donetime + exinst->delay);

  exinst->donecycle = donetime + exinst->delay;
  if (exinst->wp || current->wrong_path) {
    if (current->writer_cp[exinst->dest] == exinst->id) {
      current->writer_cp[exinst->dest] = NONE;
      current->lasthazard_cp[exinst->dest] = exinst->donecycle;
    }
  }
  if (current->follow_sync) {
    if (current->writer_sync_cp[exinst->dest] == exinst->id) {
      current->writer_sync_cp[exinst->dest] = NONE;
      current->lasthazard_sync_cp[exinst->dest] = exinst->donecycle;
    } 
  }
  if (current->writer[exinst->dest] == exinst->id) {
    current->writer[exinst->dest] = NONE;
    current->lasthazard[exinst->dest] = exinst->donecycle;
  }
  for (i=0;i<exinst->numwaiting;i++) {
    waitinst = exinst->waiter[i];
    exinst->waiter[i] = 0;
    if (waitinst->hazcycle < exinst->donecycle) {
      waitinst->hazcycle = exinst->donecycle;
    }
    if (--waitinst->deps == 0) {
      waitinst->status &= ~BLOCKED;
      waitinst->readycycle = waitinst->hazcycle;
    }
  }
  exinst->numwaiting = 0;
}

static int mem_acc_in_q[MAXTHREADS];
long mem_conflicts[4096];
int mem_con_thds[4096];

/* The main action of the queue stage is just searching through
 *  the two instruction queues for instructions to issue.
 *  Unfortunately there is no real shortcut to a linear search
 *  here which makes this a particularly slow part of the code.
 */

void
queue()
{
  scoreboard_t *new;
  int intissue=0, fpissue=0, ldstissue=0, synchissue=0;
  int i;

      for (i=0;i<numcontexts;i++) {
	mem_acc_in_q[i] = 0;
      }
      new = iq_head;
      iq_tail = NULL;
      while (new != NULL) {
	/* recall--new->readycycle is 0 unless the instruction's
	   dependencies have all been resolved.  There are an awful
	   lot of other reasons that an instruction may not be 
	   issuable, however.
	*/
	if (new->readycycle && new->readycycle <= cyc && intissue < MAXINT &&
	    !new->wait_sync &&
	    !(new->fu == INTLDST && (ldstissue >= MAXLDST)) &&
	    !(new->fu == SYNCH && (synchissue >= MAXSYNCH || 
	                           mem_acc_in_q[new->thread]))
	    && !((new->srcmem) 
		 && (mem_conflicts[(new->srcmem>>2) & 4095] == cyc)
		 && (mem_con_thds[(new->srcmem>>2) & 4095] & 
		     (1<<new->thread)))) {
	  numiniq--;
	  if (iq_tail == NULL) {
	    iq_head = new->next;
	  }
	  else {
	    iq_tail->next = new->next;
	  }
	  intissue++;
	  if (new->fu == INTLDST)
	    ldstissue++;
	  if (new->fu == SYNCH)
	    synchissue++;
#ifdef PRIO_INST_COUNT
	  key[new->thread]--;
#else
#ifdef PRIO_BR_COUNT
	  if (new->branch || new->cond_branch) {
	    key[new->thread]--;
	  }
#endif
#endif
	  if (new->wp) wpexec++;
	  if (new->branch)
	    brresolve(new);
	  else if (new->fu == SYNCH) {
	    synchresolve(new);
	  }
	  else
	    resolve(new);
	  insert_in_rread(new);
	  new = new->next;
	}
	else { /* new not scheduled */
	  /*expensive stats -- think about commenting out */
	  if (new->readycycle && new->readycycle <= cyc && !new->wait_sync &&
	      !(new->fu == SYNCH && synchissue >= MAXSYNCH)
	      && !((new->srcmem) 
		 && (mem_conflicts[(new->srcmem>>2) & 4095] == cyc)
		   && (mem_con_thds[(new->srcmem>>2) & 4095] & 
		       (1<<new->thread)))) {
	    if (intissue >= MAXINT)
	      intfuconf++;
	    if (new->fu == INTLDST && ldstissue >= MAXLDST)
	      ldstfuconf++;
	  }
	  /* establish an implied memory barrier to use in scheduling
	     release operations */
	  if (new->destmem || new->srcmem)
	    mem_acc_in_q[new->thread] = 1;
	  iq_tail = new;
	  if (new->destmem) {
	    if (mem_conflicts[((new->destmem>>2) & 4095)] == cyc) {
	      mem_con_thds[(new->destmem>>2) & 4095] |= 1<<new->thread;
	    }
	    else 
	       mem_con_thds[(new->destmem>>2) & 4095] = 1<<new->thread;
	    mem_conflicts[((new->destmem>>2) & 4095)] = cyc;
	  }
	  new = new->next;
	}
      }


  /* do pretty much the same things for the fp queue */

      new = fq_head;
      fq_tail = NULL;
      while (new != NULL) {
	if (fpissue < MAXFP && new->readycycle && new->readycycle <= cyc) {
	  if (fq_tail == NULL) {
	    fq_head = new->next;
	  }
	  else {
	    fq_tail->next = new->next;
	  }
	  fpissue++;
	  numinfq--;
#ifdef PRIO_INST_COUNT
	  key[new->thread]--;
#else
#ifdef PRIO_BR_COUNT
	  if (new->branch || new->cond_branch) {
	    key[new->thread]--;
	  }
#endif
#endif
	  if (new->wp) wpexec++;
	  if (new->branch)
	    brresolve(new);
	  else
	    resolve(new); 
	  insert_in_rread(new);
	  new = new->next;
	}
	else { /* not scheduled */
	  if (new->readycycle && new->readycycle <= cyc) {
	    if (fpissue >= MAXFP)
	      fpfuconf++;
	  }
	  fq_tail = new;
	  new = new->next;
	}
      }

      if (rread_tail) rread_tail->next = NULL;

      instrcount[INTEGER] += intissue - ldstissue - synchissue;
      instrcount[SYNCH] += synchissue;
      instrcount[INTLDST] += ldstissue;
      instrcount[FP] += fpissue;
}

