/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#include <stdio.h>
#include "filehdr.h"
#include "a.out.h"
/* #include <ldfcn.h> */
#include "main.h"
#include "mem.h"
#include <filehdr.h>
#include <syms.h>
#include <ldfcn.h>

/*
 * format of bootable a.out file headers
 */
struct execinfo {
        struct filehdr fh;
        AOUTHDR ah;
};
extern int errno;

/* Read an Alpha executable file.  Gets various address information,
   puts the executable in memory.
   Return zero if successful, else non-zero.
 */

extern int readaout(context *current, char *filename, 
		    context *previous, int share, int numthreads) {
    struct execinfo ei;
    FILE *fd;
    
    if((fd = fopen(filename, "r")) == NULL)
	error(TopLevel,"Cannot open file: %s\n", filename);
    /* read the fileheader and a.out header in one fell swoop */
    if (fread(&ei, 1, sizeof(ei), fd) != sizeof(ei)) {
	fclose(fd); error(TopLevel, "Bad header\n");
    }
    if((ei.fh.f_magic != ALPHAMAGIC) && (ei.fh.f_magic != ALPHAUMAGIC)) {
	fclose(fd); error(TopLevel, "File is not a binary executable\n");
    }
    current->pbreak = ROUND(ei.ah.bss_start + ei.ah.bsize,PAGESIZE); /* program break */
    current->pstaddress = ei.ah.entry; /* program starting address */
    current->R[REG_GP] = ei.ah.gp_value;
    DEBUGPRINTF("text_start: 0x%lx    text_size: 0x%x\n", 
		ei.ah.text_start, ei.ah.tsize);
    DEBUGPRINTF("data_start: 0x%lx    data_size: 0x%x\n", 
		ei.ah.data_start, ei.ah.dsize);
    DEBUGPRINTF("bss_start: 0x%lx    bss_size: 0x%x\n", 
		ei.ah.bss_start, ei.ah.bsize);
    DEBUGPRINTF("entry point: 0x%lx\n", ei.ah.entry);
    if (share) {
      DEBUGPRINTF("Current context 0x%lx sharing memory with context 0x%lx\n",
		  current, previous);
      share_memory(current,previous,ei.ah.tsize + 0x4000,
		   ei.ah.text_start,ei.ah.data_start, current->id);
    } else { 
      DEBUGPRINTF("Current context 0x%lx creating new memory\n", current);
      make_memory(current,
		  ei.ah.tsize + 0x4000, current->pbreak - ei.ah.data_start,
		  ei.ah.text_start,ei.ah.data_start, numthreads);
    }
    if (!share) {
      fseek(fd, N_TXTOFF(ei.fh, ei.ah), 0);
      if(fread((void *) mem_address(current, ei.ah.text_start), 1, ei.ah.tsize, fd) != 
	 ei.ah.tsize) {
	fclose(fd); 
	error(TopLevel, "Cannot read entire text segment: %s\n", filename);
      }
      if(fread((void *) mem_address(current, ei.ah.data_start), 1, ei.ah.dsize, fd) != 
	 ei.ah.dsize) {
	fclose(fd);
	error(TopLevel, "Cannot read entire data segment: %s\n", filename);
      }
    }
    bzero((void*) mem_address(current, ei.ah.bss_start), ei.ah.bsize);
    fclose(fd);
    return 0;
}
