/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

 */

#define _IEEE 1
#include <math.h>
#include "stash.h"
#include "main.h"
#define RUNC

#define THREADLENGTH 300000000

unsigned long cyc;
extern unsigned long totalsched;
extern int numcontexts;
extern int simulating;
long allinstructions;
extern context *thecontexts[];

/*
 *  run() is the controller for virtually all simulation.
 *
 *  Each cycle it simulates each pipeline stage, then does some
 *    accounting and mop-up before advancing the cycle counter.
 */

int run()
{

    unsigned long statpoint = (long) 1000000;

/*  Because I simulate for a constant number of TOTAL instructions,
     optimizations that subtly change the relative priority of
     threads can change (e.g., improve) reported performance without
     actually doing anything positive to the system.  I watch that
     carefully.  You should too.
 */
    allinstructions = (long) numcontexts * (long) THREADLENGTH;
    simulating = 1;

    while(1)
    {
      /*  going through the pipeline back to front ensures that instructions
	  move forward through the pipe accurately (eg, they get backed up if
	  later stages stall), but the danger is that information that moves
	  backward through the pipeline moves too quickly (ie, in the same
	  cycle).  To combat this, the fix_pcs() routine doesn't correct the
	  fetch address on mispredicts and misfetches until after the fetch
	  stage.  Also, fix_regs() ensures that rename registers aren't reused
	  the same cycle */

	commit();
	regwrite();
	execute();
	regread();
	queue();
	regrename();
	decode();
	fetch();
	fix_pcs();
	fix_regs();
	calculate_priority();/* fetch priority */
	process_cache_queues();/* handle all memory events*/
	cyc++;
	DEBUGPRINTF("cyc = %ld\n",cyc);
/*
 *  I print out stats regularly so I can monitor progress, predict
 *    finish times, and catch runaway simulations.  Delete this code
 *    if you don't want that.
 */
      if (cyc >= statpoint) {
	statpoint += (long) 100000000;
	schedstats();
	print_cstats();
	predict_stats();
	fflush(0);
      }

#ifndef LONG
     /* in this case, don't stop until someone exits */
      if ( (long) allinstructions < (long) 0) {
	printf("***** exiting *****\n");
	schedstats();
	print_cstats();
	predict_stats();
	fflush(0);
	exit(0);
      }
#endif
    }
}
