/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#include <stdlib.h>
#include "main.h"
#include "mem.h"
#include "stack.h"

#define TEXT_START 0x20000000
StackChunk stackspace[STACKSPACESZ];

StackChunk *freelist = NULL;

void
init_stack_space() {
  int i;
  for (i=0;i<STACKSPACESZ;i++) {
    stackspace[i].number = i;
    stackspace[i].free = 1;
    stackspace[i].contextid = -1;
  }
  stackspace[STACKSPACESZ-1].base = (mem_addr)TEXT_START - STACKCHUNKSZ;
  for (i=STACKSPACESZ-2; i >=0; i--) {
    stackspace[i].base = stackspace[i+1].base - STACKCHUNKSZ; 
  }
  freelist = &stackspace[STACKSPACESZ-1];
}

mem_addr
stack_allocate_contiguous(context *current, unsigned long size) {
  int numchunks;
  int start, curr;
  int consecutive, prevnumber, i;
  
  numchunks = (int)size/STACKCHUNKSZ;
  if (size % STACKCHUNKSZ)
    numchunks++;

  DEBUGPRINTF("stack_allocating %d contiguous chunk(s) for desired stack size %ld\n",
	      numchunks, size);
  if (!freelist) {
    fprintf(stderr,"stack_allocate_contiguous failed:\n"); 
    fprintf(stderr, "No free stack chunks left!\n");
    return NULL;
  }
  prevnumber = freelist->number;
  for(start = freelist->number, curr = freelist->number-1, consecutive = 1;
      (curr >= 0) && (consecutive < numchunks);
      curr--) {
    if (stackspace[curr].free) {
      consecutive++;
      prevnumber = curr;
    } else {
      consecutive = 0;
      start = curr;
    }
  }
  if (consecutive == numchunks) {
    for (i = consecutive-1; i >= 0; i--) {
      stackspace[start].free = 0;
      stackspace[start].contextid = current->id;
      start--;
    }
    for (i = start; i >=0; i--) {
      if (stackspace[i].free) {
	freelist = &stackspace[i];
	DEBUGPRINTF("Freelist now begins at chunk %d\n", i);
	break;
      }
    }
    if (i < 0) {
      freelist = NULL;
      DEBUGPRINTF("No more free chunks\n");
    }
      
    start++;

    DEBUGPRINTF("\tFirst chunk at %d, base = 0x%lx\n", start, 
		stackspace[start].base);
    return stackspace[start].base;
  }
  fprintf(stderr, "Couldn't find %d contiguous stack chunks\n",
	  numchunks);
  return NULL;
}

void
stack_free(mem_addr base) {
  int i;

  int start_number;

  for (i = 0; i < STACKSPACESZ; i++) {
    if (stackspace[i].base == base)
      break;
  }
  if (i == STACKSPACESZ)
    fprintf(stderr, "Couldn't find stack based at 0x%lx\n", base);

  for (start_number = i; 
       (stackspace[start_number].contextid == stackspace[i].contextid);
       i++) {
    stackspace[i].free = 1;
    stackspace[i].contextid = -1;
  }

  if (start_number < freelist->number)
    freelist = &stackspace[start_number];
}
